#!/bin/bash
########################################################################
# build_gojsonnet.sh
#
# Build gojsonnet and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/bin"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=gojsonnet
origpkgver=v0_18_0
pkgver=${origpkgver} # Append letters for local patches.
ssibuildshims_version=v2_16_03
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${pkgdotver}
pkgtarfile=${pkgtarname}.tar
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,e26,c7,c14,c15} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_BUILD

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  WANT_TESTS

    Activate tests.

EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics IN_SOURCE NO_TYPE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
${trace} setup_prod golang --latest || exit
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" "${tardir}/${pkgtarfile}" || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

ncores=$(ncores)

cxx_opt_flags=()
case $cqual in
  e19|e20|e26) cc=gcc; cxx=g++; cxxstd=17;;
  c7|c14|c15) cc=clang; cxx=clang++; cxxstd=17
      cxx_opt_flags+=(-D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1);;
  *) ssi_die -I "accepted unknown cqual=$cqual"
esac

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}/${pkgtarname}" || \
  ssi_die "unable to cd to build directory ${blddir}"

# Make sure things under $blddir can be deleted after we exit.
trap "[ -d \"$blddir\" ] && chmod -R u+w \"$blddir\"" EXIT

##################
# Build and install jsonnet and jsonnetfmt

export CC=$cc
export CXX=$cxx
export CGO_CXXFLAGS="-std=c++$cxxstd${cxx_opt_flags[*]:+ ${cxx_opt_flags[*]}}"
export GOPATH=${pkgdir}
echo "GOBIN variable: " ${GOBIN}
${trace} mkdir -p $GOPATH/bin || ssi_die "could not create $GOPATH/bin subdireectory"
${trace} go install -p $ncores -x ./cmd/jsonnet || ssi_die "build of jsonnet failed"
${trace} go install -p $ncores -x ./cmd/jsonnetfmt || ssi_die "build of jsonnetfmt failed"

# Build c-bindings
${trace} cd c-bindings || \
  ssi_die "unable to cd to build c-bindings subdirectory"

ldflags=()
if [ "$OS1" = Darwin ]; then
  ldflags+=(-extldflags=-Wl,--unresolved-symbols=ignore-in-shared-libs)
else
  ldflags+=(-extldflags=-Wl,-undefined,error)
fi
${trace} go build -p $ncores -buildmode=c-shared \
         ${ldflags[*]:+-ldflags "${ldflags[*]}"} \
         -x -o libjsonnet.so || \
  ssi_die "build of jsonnet c-bindings failed"

##################
# Adjust the generated libjsonnet.h file, which is missing one of the C API functions
# -- This is icky, icky (see https://github.com/google/go-jsonnet/issues/562)
(
  ${trace} patch -p1 << EOF
--- a/libjsonnet.h	2023-03-15 11:28:27.285196486 -0500
+++ b/libjsonnet.h	2023-03-15 11:37:32.164098110 -0500
@@ -105,6 +105,7 @@
 extern struct JsonnetJsonValue* jsonnet_json_make_object(struct JsonnetVm* vmRef);
 extern void jsonnet_json_object_append(struct JsonnetVm* vmRef, struct JsonnetJsonValue* obj, char* f, struct JsonnetJsonValue* v);
 extern void jsonnet_json_destroy(struct JsonnetVm* vmRef, struct JsonnetJsonValue* v);
+extern char* jsonnet_realloc(struct JsonnetVm* vm, char* str, size_t sz);
 extern void jsonnet_fmt_indent(struct JsonnetVm* vmRef, int n);
 extern void jsonnet_fmt_max_blank_lines(struct JsonnetVm* vmRef, int n);
 extern void jsonnet_fmt_string(struct JsonnetVm* vmRef, int c);
EOF
) || ssi_die "unable to adjust libjsonnet.h"

ssi_info  "finished patching libjsonnet.h for ${package} ${pkgver}"
##################
# Install c-bindings by hand

${trace} mkdir -p ${pkgdir}/{include/gojsonnet,lib} || ssi_die "could not create ${pkgdir}/{include,lib} subdirectories"
${trace} install *.so ${pkgdir}/lib || ssi_die "could not install libraries"
${trace} sed -Ee 's&([\"<])(internal\.h[\">])$&\1gojsonnet/\2&' libjsonnet.h > "$pkgdir/include/libjsonnet.h" && \
  ${trace} install -m 644 internal.h "$pkgdir/include/gojsonnet/" || \
    ssi_die "unable to install headers"
##################

# Make sure everything installed can be deleted by the owner:
${trace} chmod -R u+w "$pkgdir" >/dev/null 2>&1

ssi_info "finished building ${package} ${pkgver}"

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
